#include "css-colors.h"
const CSSColor CSSColors[] = {
    {.name = "AliceBlue", .r = 0xF0, .g = 0xF8, .b = 0xFF},
    {.name = "AntiqueWhite", .r = 0xFA, .g = 0xEB, .b = 0xD7},
    {.name = "Aqua", .r = 0x00, .g = 0xFF, .b = 0xFF},
    {.name = "Aquamarine", .r = 0x7F, .g = 0xFF, .b = 0xD4},
    {.name = "Azure", .r = 0xF0, .g = 0xFF, .b = 0xFF},
    {.name = "Beige", .r = 0xF5, .g = 0xF5, .b = 0xDC},
    {.name = "Bisque", .r = 0xFF, .g = 0xE4, .b = 0xC4},
    {.name = "Black", .r = 0x00, .g = 0x00, .b = 0x00},
    {.name = "BlanchedAlmond", .r = 0xFF, .g = 0xEB, .b = 0xCD},
    {.name = "Blue", .r = 0x00, .g = 0x00, .b = 0xFF},
    {.name = "BlueViolet", .r = 0x8A, .g = 0x2B, .b = 0xE2},
    {.name = "Brown", .r = 0xA5, .g = 0x2A, .b = 0x2A},
    {.name = "BurlyWood", .r = 0xDE, .g = 0xB8, .b = 0x87},
    {.name = "CadetBlue", .r = 0x5F, .g = 0x9E, .b = 0xA0},
    {.name = "Chartreuse", .r = 0x7F, .g = 0xFF, .b = 0x00},
    {.name = "Chocolate", .r = 0xD2, .g = 0x69, .b = 0x1E},
    {.name = "Coral", .r = 0xFF, .g = 0x7F, .b = 0x50},
    {.name = "CornflowerBlue", .r = 0x64, .g = 0x95, .b = 0xED},
    {.name = "Cornsilk", .r = 0xFF, .g = 0xF8, .b = 0xDC},
    {.name = "Crimson", .r = 0xDC, .g = 0x14, .b = 0x3C},
    {.name = "Cyan", .r = 0x00, .g = 0xFF, .b = 0xFF},
    {.name = "DarkBlue", .r = 0x00, .g = 0x00, .b = 0x8B},
    {.name = "DarkCyan", .r = 0x00, .g = 0x8B, .b = 0x8B},
    {.name = "DarkGoldenRod", .r = 0xB8, .g = 0x86, .b = 0x0B},
    {.name = "DarkGray", .r = 0xA9, .g = 0xA9, .b = 0xA9},
    {.name = "DarkGrey", .r = 0xA9, .g = 0xA9, .b = 0xA9},
    {.name = "DarkGreen", .r = 0x00, .g = 0x64, .b = 0x00},
    {.name = "DarkKhaki", .r = 0xBD, .g = 0xB7, .b = 0x6B},
    {.name = "DarkMagenta", .r = 0x8B, .g = 0x00, .b = 0x8B},
    {.name = "DarkOliveGreen", .r = 0x55, .g = 0x6B, .b = 0x2F},
    {.name = "DarkOrange", .r = 0xFF, .g = 0x8C, .b = 0x00},
    {.name = "DarkOrchid", .r = 0x99, .g = 0x32, .b = 0xCC},
    {.name = "DarkRed", .r = 0x8B, .g = 0x00, .b = 0x00},
    {.name = "DarkSalmon", .r = 0xE9, .g = 0x96, .b = 0x7A},
    {.name = "DarkSeaGreen", .r = 0x8F, .g = 0xBC, .b = 0x8F},
    {.name = "DarkSlateBlue", .r = 0x48, .g = 0x3D, .b = 0x8B},
    {.name = "DarkSlateGray", .r = 0x2F, .g = 0x4F, .b = 0x4F},
    {.name = "DarkSlateGrey", .r = 0x2F, .g = 0x4F, .b = 0x4F},
    {.name = "DarkTurquoise", .r = 0x00, .g = 0xCE, .b = 0xD1},
    {.name = "DarkViolet", .r = 0x94, .g = 0x00, .b = 0xD3},
    {.name = "DeepPink", .r = 0xFF, .g = 0x14, .b = 0x93},
    {.name = "DeepSkyBlue", .r = 0x00, .g = 0xBF, .b = 0xFF},
    {.name = "DimGray", .r = 0x69, .g = 0x69, .b = 0x69},
    {.name = "DimGrey", .r = 0x69, .g = 0x69, .b = 0x69},
    {.name = "DodgerBlue", .r = 0x1E, .g = 0x90, .b = 0xFF},
    {.name = "FireBrick", .r = 0xB2, .g = 0x22, .b = 0x22},
    {.name = "FloralWhite", .r = 0xFF, .g = 0xFA, .b = 0xF0},
    {.name = "ForestGreen", .r = 0x22, .g = 0x8B, .b = 0x22},
    {.name = "Fuchsia", .r = 0xFF, .g = 0x00, .b = 0xFF},
    {.name = "Gainsboro", .r = 0xDC, .g = 0xDC, .b = 0xDC},
    {.name = "GhostWhite", .r = 0xF8, .g = 0xF8, .b = 0xFF},
    {.name = "Gold", .r = 0xFF, .g = 0xD7, .b = 0x00},
    {.name = "GoldenRod", .r = 0xDA, .g = 0xA5, .b = 0x20},
    {.name = "Gray", .r = 0x80, .g = 0x80, .b = 0x80},
    {.name = "Grey", .r = 0x80, .g = 0x80, .b = 0x80},
    {.name = "Green", .r = 0x00, .g = 0x80, .b = 0x00},
    {.name = "GreenYellow", .r = 0xAD, .g = 0xFF, .b = 0x2F},
    {.name = "HoneyDew", .r = 0xF0, .g = 0xFF, .b = 0xF0},
    {.name = "HotPink", .r = 0xFF, .g = 0x69, .b = 0xB4},
    {.name = "IndianRed", .r = 0xCD, .g = 0x5C, .b = 0x5C},
    {.name = "Indigo", .r = 0x4B, .g = 0x00, .b = 0x82},
    {.name = "Ivory", .r = 0xFF, .g = 0xFF, .b = 0xF0},
    {.name = "Khaki", .r = 0xF0, .g = 0xE6, .b = 0x8C},
    {.name = "Lavender", .r = 0xE6, .g = 0xE6, .b = 0xFA},
    {.name = "LavenderBlush", .r = 0xFF, .g = 0xF0, .b = 0xF5},
    {.name = "LawnGreen", .r = 0x7C, .g = 0xFC, .b = 0x00},
    {.name = "LemonChiffon", .r = 0xFF, .g = 0xFA, .b = 0xCD},
    {.name = "LightBlue", .r = 0xAD, .g = 0xD8, .b = 0xE6},
    {.name = "LightCoral", .r = 0xF0, .g = 0x80, .b = 0x80},
    {.name = "LightCyan", .r = 0xE0, .g = 0xFF, .b = 0xFF},
    {.name = "LightGoldenRodYellow", .r = 0xFA, .g = 0xFA, .b = 0xD2},
    {.name = "LightGray", .r = 0xD3, .g = 0xD3, .b = 0xD3},
    {.name = "LightGrey", .r = 0xD3, .g = 0xD3, .b = 0xD3},
    {.name = "LightGreen", .r = 0x90, .g = 0xEE, .b = 0x90},
    {.name = "LightPink", .r = 0xFF, .g = 0xB6, .b = 0xC1},
    {.name = "LightSalmon", .r = 0xFF, .g = 0xA0, .b = 0x7A},
    {.name = "LightSeaGreen", .r = 0x20, .g = 0xB2, .b = 0xAA},
    {.name = "LightSkyBlue", .r = 0x87, .g = 0xCE, .b = 0xFA},
    {.name = "LightSlateGray", .r = 0x77, .g = 0x88, .b = 0x99},
    {.name = "LightSlateGrey", .r = 0x77, .g = 0x88, .b = 0x99},
    {.name = "LightSteelBlue", .r = 0xB0, .g = 0xC4, .b = 0xDE},
    {.name = "LightYellow", .r = 0xFF, .g = 0xFF, .b = 0xE0},
    {.name = "Lime", .r = 0x00, .g = 0xFF, .b = 0x00},
    {.name = "LimeGreen", .r = 0x32, .g = 0xCD, .b = 0x32},
    {.name = "Linen", .r = 0xFA, .g = 0xF0, .b = 0xE6},
    {.name = "Magenta", .r = 0xFF, .g = 0x00, .b = 0xFF},
    {.name = "Maroon", .r = 0x80, .g = 0x00, .b = 0x00},
    {.name = "MediumAquaMarine", .r = 0x66, .g = 0xCD, .b = 0xAA},
    {.name = "MediumBlue", .r = 0x00, .g = 0x00, .b = 0xCD},
    {.name = "MediumOrchid", .r = 0xBA, .g = 0x55, .b = 0xD3},
    {.name = "MediumPurple", .r = 0x93, .g = 0x70, .b = 0xDB},
    {.name = "MediumSeaGreen", .r = 0x3C, .g = 0xB3, .b = 0x71},
    {.name = "MediumSlateBlue", .r = 0x7B, .g = 0x68, .b = 0xEE},
    {.name = "MediumSpringGreen", .r = 0x00, .g = 0xFA, .b = 0x9A},
    {.name = "MediumTurquoise", .r = 0x48, .g = 0xD1, .b = 0xCC},
    {.name = "MediumVioletRed", .r = 0xC7, .g = 0x15, .b = 0x85},
    {.name = "MidnightBlue", .r = 0x19, .g = 0x19, .b = 0x70},
    {.name = "MintCream", .r = 0xF5, .g = 0xFF, .b = 0xFA},
    {.name = "MistyRose", .r = 0xFF, .g = 0xE4, .b = 0xE1},
    {.name = "Moccasin", .r = 0xFF, .g = 0xE4, .b = 0xB5},
    {.name = "NavajoWhite", .r = 0xFF, .g = 0xDE, .b = 0xAD},
    {.name = "Navy", .r = 0x00, .g = 0x00, .b = 0x80},
    {.name = "OldLace", .r = 0xFD, .g = 0xF5, .b = 0xE6},
    {.name = "Olive", .r = 0x80, .g = 0x80, .b = 0x00},
    {.name = "OliveDrab", .r = 0x6B, .g = 0x8E, .b = 0x23},
    {.name = "Orange", .r = 0xFF, .g = 0xA5, .b = 0x00},
    {.name = "OrangeRed", .r = 0xFF, .g = 0x45, .b = 0x00},
    {.name = "Orchid", .r = 0xDA, .g = 0x70, .b = 0xD6},
    {.name = "PaleGoldenRod", .r = 0xEE, .g = 0xE8, .b = 0xAA},
    {.name = "PaleGreen", .r = 0x98, .g = 0xFB, .b = 0x98},
    {.name = "PaleTurquoise", .r = 0xAF, .g = 0xEE, .b = 0xEE},
    {.name = "PaleVioletRed", .r = 0xDB, .g = 0x70, .b = 0x93},
    {.name = "PapayaWhip", .r = 0xFF, .g = 0xEF, .b = 0xD5},
    {.name = "PeachPuff", .r = 0xFF, .g = 0xDA, .b = 0xB9},
    {.name = "Peru", .r = 0xCD, .g = 0x85, .b = 0x3F},
    {.name = "Pink", .r = 0xFF, .g = 0xC0, .b = 0xCB},
    {.name = "Plum", .r = 0xDD, .g = 0xA0, .b = 0xDD},
    {.name = "PowderBlue", .r = 0xB0, .g = 0xE0, .b = 0xE6},
    {.name = "Purple", .r = 0x80, .g = 0x00, .b = 0x80},
    {.name = "RebeccaPurple", .r = 0x66, .g = 0x33, .b = 0x99},
    {.name = "Red", .r = 0xFF, .g = 0x00, .b = 0x00},
    {.name = "RosyBrown", .r = 0xBC, .g = 0x8F, .b = 0x8F},
    {.name = "RoyalBlue", .r = 0x41, .g = 0x69, .b = 0xE1},
    {.name = "SaddleBrown", .r = 0x8B, .g = 0x45, .b = 0x13},
    {.name = "Salmon", .r = 0xFA, .g = 0x80, .b = 0x72},
    {.name = "SandyBrown", .r = 0xF4, .g = 0xA4, .b = 0x60},
    {.name = "SeaGreen", .r = 0x2E, .g = 0x8B, .b = 0x57},
    {.name = "SeaShell", .r = 0xFF, .g = 0xF5, .b = 0xEE},
    {.name = "Sienna", .r = 0xA0, .g = 0x52, .b = 0x2D},
    {.name = "Silver", .r = 0xC0, .g = 0xC0, .b = 0xC0},
    {.name = "SkyBlue", .r = 0x87, .g = 0xCE, .b = 0xEB},
    {.name = "SlateBlue", .r = 0x6A, .g = 0x5A, .b = 0xCD},
    {.name = "SlateGray", .r = 0x70, .g = 0x80, .b = 0x90},
    {.name = "SlateGrey", .r = 0x70, .g = 0x80, .b = 0x90},
    {.name = "Snow", .r = 0xFF, .g = 0xFA, .b = 0xFA},
    {.name = "SpringGreen", .r = 0x00, .g = 0xFF, .b = 0x7F},
    {.name = "SteelBlue", .r = 0x46, .g = 0x82, .b = 0xB4},
    {.name = "Tan", .r = 0xD2, .g = 0xB4, .b = 0x8C},
    {.name = "Teal", .r = 0x00, .g = 0x80, .b = 0x80},
    {.name = "Thistle", .r = 0xD8, .g = 0xBF, .b = 0xD8},
    {.name = "Tomato", .r = 0xFF, .g = 0x63, .b = 0x47},
    {.name = "Turquoise", .r = 0x40, .g = 0xE0, .b = 0xD0},
    {.name = "Violet", .r = 0xEE, .g = 0x82, .b = 0xEE},
    {.name = "Wheat", .r = 0xF5, .g = 0xDE, .b = 0xB3},
    {.name = "White", .r = 0xFF, .g = 0xFF, .b = 0xFF},
    {.name = "WhiteSmoke", .r = 0xF5, .g = 0xF5, .b = 0xF5},
    {.name = "Yellow", .r = 0xFF, .g = 0xFF, .b = 0x00},
    {.name = "YellowGreen", .r = 0x9A, .g = 0xCD, .b = 0x32}};

const unsigned int num_CSSColors = sizeof(CSSColors) / sizeof(*CSSColors);
