#include "rofi-types.h"

/**
 * Name of the property type
 */
const char *const PropertyTypeName[P_NUM_TYPES] = {
    /** Integer */
    "Integer",
    /** Double */
    "Double",
    /** String */
    "String",
    /** Boolean */
    "Boolean",
    /** Color */
    "Color",
    /** Image */
    "Image",
    /** Padding */
    "Padding",
    /** Link to global setting */
    "Reference",
    /** Position */
    "Position",
    /** Highlight */
    "Highlight",
    /** List */
    "List",
    /** Orientation */
    "Orientation",
    /** Cursor */
    "Cursor",
    /** Inherit */
    "Inherit",
};
